package io.takima.agencymanagement.service;

import io.takima.agencymanagement.dao.TravelDao;
import io.takima.agencymanagement.exception.AlreadyTravelSubscribedException;
import io.takima.agencymanagement.mapper.TravelDtoMapper;
import io.takima.agencymanagement.model.Travel;
import io.takima.agencymanagement.model.User;
import io.takima.agencymanagement.presentation.responsedto.TravelResponseDto;
import io.takima.agencymanagement.model.Discount;

import java.time.Duration;
import java.time.Instant;
import java.util.Comparator;
import java.util.List;
import java.util.NoSuchElementException;

import static io.takima.agencymanagement.mapper.TravelDtoMapper.fromTravel;

public class TravelService {

    private TravelDao travelDao;

    public TravelService() {
        travelDao = new TravelDao();
    }

    public TravelResponseDto create(Travel travel) {
        travelDao.save(travel);
        return fromTravel(travel);
    }

    public TravelResponseDto update(Travel travel) {
        travelDao.save(travel);
        return fromTravel(travel);
    }

    public TravelResponseDto delete(Travel travel) {
        travelDao.save(travel);
        return fromTravel(travel);
    }

    public TravelResponseDto findById(Long id) {
        Travel travel = travelDao.findById(id);
        return fromTravel(travel);
    }

    public List<TravelResponseDto> findAll() {
        return travelDao.findAll().stream()
                .map(TravelDtoMapper::fromTravel)
                .toList();

    }

    public void applyDiscounts(Travel travel) {
        if (travel.getDiscounts().isEmpty()) {
            System.out.println("Aucune réduction n'est applicable sur ce voyage !");
        } else {
            double discountedPrice = travel.getPrice();
            for (Discount discount : travel.getDiscounts()) {
                discountedPrice = discount.apply(discountedPrice);
            }
            System.out.println("Prix après les réductions : " + discountedPrice);
        }
    }

    public void subscribe(User user, Travel travel) throws Exception {

        if (travel.getParticipants().contains(user) || travel.getWaitList().contains(user)) {
            //User already subscribed or in waitList
            throw new AlreadyTravelSubscribedException("L'utilisateur est déjà inscrit ou en file d'attente pour ce voyage.");
        }

        if (travel.getParticipants().size() < travel.getCapacity()) {
            travel.getParticipants().add(user);
            return;
        }
        travel.getWaitList().add(user);
    }

    public void unsubscribe(User user, Travel travel) {

        travel.getParticipants().remove(user);

        if (!travel.getWaitList().isEmpty()) {
            travel.getParticipants().add(travel.getWaitList().poll());
        }
    }

    public Duration computeTravelDuration(Travel travel) {
        return Duration.between(travel.getDepartureDate(), travel.getArrivalDate());
    }

    public void adjustDates(Duration delay, Travel travel) {
        travel.setDepartureDate(travel.getDepartureDate().plus(delay));
        travel.setArrivalDate(travel.getArrivalDate().plus(delay));
    }

    public Travel findClosestToCurrentDate(List<Travel> travels) {
        Instant now = Instant.now();

        return travels.stream()
                .min(Comparator.comparing(travel -> Duration.between(now, travel.getDepartureDate()).abs()))
                .orElseThrow(NoSuchElementException::new);
    }

    public List<Travel> findAvailableForDestination(String destination) {
        List<Travel> travels = travelDao.findAll();

        return travels.stream()
                .filter(travel -> travel.hasAvailableCapacity(travel.getParticipants(), travel.getCapacity()) && travel.getDestination().equals(destination))
                .toList();
    }


    public List<Travel> findInPriceRange(double minPrice, double maxPrice) {
        List<Travel> travels = travelDao.findAll();

        return travels.stream()
                .filter(travel -> travel.getPrice() >= minPrice && travel.getPrice() <= maxPrice)
                .toList();
    }

    public List<Travel> findNSortedByPriceAsc(int n) {
        List<Travel> travels = travelDao.findAll();

        return travels.stream()
                .sorted(Comparator.comparing(Travel::getPrice))
                .limit(n)
                .toList();
    }

    public List<Travel> findWithPromotions() {
        List<Travel> travels = travelDao.findAll();

        return travels.stream()
                .filter(travel -> !travel.getDiscounts().isEmpty())
                .toList();
    }

    public List<Travel> findByUser(String firstName, String lastName) {
        List<Travel> travels = travelDao.findAll();

        return travels.stream()
                .filter(travel ->
                        travel.getParticipants().stream()
                                .anyMatch(user ->
                                        user.getFirstName().equals(firstName) &&
                                                user.getLastName().equals(lastName)))
                .toList();
    }

    public double getAveragePrice() {
        List<Travel> travels = travelDao.findAll();

        return travels.stream()
                .mapToDouble(Travel::getPrice)
                .average()
                .orElse(0.0);
    }
}
